(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
  typeof define === 'function' && define.amd ? define(['exports'], factory) :
  (global = global || self, factory(global.iterablefu = {}));
}(this, function (exports) { 'use strict';

  /**
   * Dynamically create a ChainableClass. This differs from makeChainableIterator only in that the class can't be
   * called as a function.
   *
   * @param {Object} generators - Each key is the name of a generator, the value is a generator function.
   * @param {Object} transforms - Each key is the name of a transform, the value is a generator function that takes
   * an iterable as the last argument.
   * @param {Object} reducers - Each key is the name of a reducer, the value is a function that takes the iterable
   * to reduce as the last argument.
   */
  const makeChainableClass = (generators, transforms, reducers) => {
    // construct chainable iterable class using both class semantics
    // then add methods using the classic prototype assignment technique
    const Chainable = class {
      constructor (iterable) {
        this.chainedIterable = iterable;
      }

      * [Symbol.iterator] () {
        yield * this.chainedIterable;
      }
    };

    // Dynamically add static Sequence methods to class
    for (const methodName in generators) {
      Chainable[methodName] = function (...args) {
        return new Chainable(generators[methodName](...args))
      };
    }

    // Dynamically add Transform methods to class
    for (const methodName in transforms) {
      Chainable.prototype[methodName] = function (...args) {
        this.chainedIterable = transforms[methodName](...args, this.chainedIterable);
        return this
      };
    }

    // Dynamically add Reducer methods to class
    for (const methodName in reducers) {
      Chainable.prototype[methodName] = function (...args) {
        return reducers[methodName](...args, this.chainedIterable)
      };
    }
    return Chainable
  };

  /**
   * Dynamically create a ChainableIterable class/function.
   *
   * @param {Object} generators - Each key is the name of a generator, the value is a generator function.
   * @param {Object} transforms - Each key is the name of a transform, the value is a generator function that takes
   * the iterable to transform as the last argument.
   * @param {Object} reducers - Each key is the name of a reducer, the value is a function that takes the iterable to
   * reduce as the last argument.
   */
  const makeChainableIterable = (generators, transforms, reducers) => {
    const ChainableClass = makeChainableClass(generators, transforms, reducers);

    const ChainableIterable = function (iterable) {
      return new ChainableClass(iterable)
    };

    ChainableIterable.ChainableIterable = ChainableClass; // provided to support testing

    // Dynamically add static sequence generator methods to class
    for (const methodName in generators) {
      ChainableIterable[methodName] = function (...args) {
        return ChainableClass[methodName](...args)
      };
    }

    return ChainableIterable
  };

  /**
   * Concatenates a list of iterables into a single iterable.
   *
   * @param  {...Iterable} iterables to be concatenated
   * @returns {Generator} that provides the output of each iterable in turn
   * @example
   * const generator = concatenate([0, 1, 2], [3, 4])
   * console.log([...generator]) // prints [0, 1, 2, 3, 4]
   */
  const concatenate = function * (...iterables) {
    for (const iterable of iterables) {
      yield * iterable;
    }
  };

  /**
   * Creates a sequence from the input sequence. This function exists solely
   * so that ChainableIterable has a static constructor method.
   *
   * @param {Iterable} inputIterable - the iterable
   * @returns {Generator} that represents the same iterable that was passed in
   * @example
   * const generatorFunction = function * (n) {
   *   for (let i = 0; i < n; i++) {
   *     yield i
   *   }
   * }
   * const generator = from(generatorFunction(3))
   * console.log([...generator]) // prints [0, 1, 2]
   */
  const from = function (iterable) {
    return iterable
  };

  /**
   * Creates a sequence of numbers similar to the Python range. See the example.
   *
   * @param  {...integer} args per the example
   * @returns {Generator} that represents the range
   * @example
   * console.log([...range()]) // prints []
   *
   * console.log([...range(5)]) // prints [0, 1, 2, 3, 4]
   *
   * console.log([...range(2, 5)]) // prints [2, 3, 4, 5, 6]
   *
   * console.log([...range(2, 5, 3)] // prints [2, 5, 8, 11, 14]
   */
  const range = (...args) => {
    const ranger = function * (start, length, increment) {
      let value = start;
      for (let i = 0; i < length; i++) {
        yield value;
        value += increment;
      }
    };
    switch (args.length) {
      case 0:
        return ranger(0, 0, 1)
      case 1:
        return ranger(0, args[0], 1)
      case 2:
        return ranger(args[0], args[1], 1)
      default:
        return ranger(...args)
    }
  };

  /**
   * Generates a sequence of things, n times
   *
   * @param {number} n - the number of times to repeat thing
   * @param {any} thing - the repeated thing
   * @returns {Generator} that will repeat thing, n times
   * @example
   * const generator = repeat(5, 'a')
   * console.log([...generator]) // prints ['a', 'a', 'a', 'a', 'a']
   */
  const repeat = function * (n, thing) {
    for (let i = 0; i < n; i++) {
      yield thing;
    }
  };

  /**
   * Repeat an iterable n times.
   *
   * NOTE: Generator functions are designed to create one-time-use iterables, and will not work as expected
   * with repeatIterable. Once a generator completes, it won't restart, and therefore can't be repeated.
   *
   * Instead, use an iterable object where calling [Symbol.iterator] returns a new Generator object with
   * new state. See the examples below...
   *
   * @param {number} n - number of times to repeat iterable
   * @param {Iterable} repeatableIterable - the input iterable to repeat, see notes above and examples.
   * @returns {Generator} - that will repeat the input iterable n times
   * @example
   * // As noted above, use iterable objects, not generator functions with repeatIterable
   * const repeatable = (length) => {
   *   return {
   *     * [Symbol.iterator] () {
   *       for (let i = 0; i < length; i++) {
   *         yield i
   *       }
   *     }
   *   }
   * }
   * const a = repeatIterable(3, repeatable(3))
   * console.log([...a]) // prints [0, 1, 2, 0, 1, 2, 0, 1, 2] as expected
   *
    * // NOTE: This generator function will not work as expected with repeatIterable.
   * const oneTime = function * (length) {
   *   for (let i = 0; i < length; i++) {
   *     yield i
   *   }
   * }
   * const b = repeatIterable(3, oneTime(3))
   * console.log([...b]) // prints [0, 1, 2] OOPS!!!!
   */
  const repeatIterable = function * (n, repeatableIterable) {
    for (let i = 0; i < n; i++) {
      yield * repeatableIterable;
    }
  };

  /**
   * Creates a sequence of arrays the same length as the *shortest* iterable provided. The first array contains the first
   * element from each of the iterables provided. The second array contains the second element from each of the
   * iterables provided, and so on.
   *
   * @param  {...Iterable} iterables the iterables to be zipped
   * @returns {Generator} for the zipped arrays
   * @example
   * const a = [0, 1, 2]
   * const b = ['a', 'b', 'c', 'd', 'e', 'f'] // note that this array is longer than a
   * const generator = zip(a, b)
   * console.log([...generator]) // prints [[0, 'a'], [1, 'b'], [2, 'c']]
   *
   */
  const zip = function * (...iterables) {
    const isDone = (nextResults) => nextResults.reduce((done, nextResult) => done || nextResult.done, false);
    const iterators = iterables.map(x => x[Symbol.iterator]());
    let nextResults = iterators.map(i => i.next());
    while (!isDone(nextResults)) {
      yield nextResults.map(result => result.value);
      nextResults = iterators.map(i => i.next());
    }
  };

  /**
   * Creates a sequence of arrays the same length as the *longest* iterable provided. The first array contains the first
   * element from each of the iterables provided. The second array contains the second element from each of the
   * iterables provided, and so on. Missing elements from the shorter iterables are set to undefined.
   *
   * @param  {...Iterable} iterables the iterables to be zipped
   * @returns {Generator} for the zipped arrays
   * @example
   * const a = [0, 1, 2]
   * const b = ['a', 'b', 'c', 'd'] // note that this array is longer than a
   * const generator = zipAll(a, b)
   * console.log([...generator]) // prints [[0, 'a'], [1, 'b'], [2, 'c'], [undefined, 'd']]
   *
   */
  const zipAll = function * (...iterables) {
    const isDone = (nextResults) => nextResults.reduce((done, nextResult) => done && nextResult.done, true);
    const iterators = iterables.map(x => x[Symbol.iterator]());
    let nextResults = iterators.map(i => i.next());
    while (!isDone(nextResults)) {
      yield nextResults.map(result => result.value);
      nextResults = iterators.map(i => i.next());
    }
  };

  var generators = /*#__PURE__*/Object.freeze({
    concatenate: concatenate,
    from: from,
    range: range,
    repeat: repeat,
    repeatIterable: repeatIterable,
    zip: zip,
    zipAll: zipAll
  });

  /**
   * Converts a sequence of Arrays to a sequence of Objects by assigning the property names
   * to each array element in turn. The input sequence doesn't have to provide arrays, it can
   * provide any sequence of iterable objects.
   *
   * If the arrays are too long, extra values are ignored.
   *
   * If the arrays are too short, the remaining properties are assigned undefined.
   *
   * @param {Iterable} propertyNames - a sequence of property names
   * @param {Iterable} iterable - a sequence of arrays (or any iterable objects)
   * @return {Generator} for the sequence of Objects
   * @example
   * const objects = arrayToObject(['a', 'b'], [[0, 1], [2, 3, 'a'], [4]])
   * // objects is [{'a': 0, 'b': 1 }, {'a': 2, 'b': 3 }, {'a': 4, 'b': undefined }]
   */
  const arrayToObject = function (propertyNames, iterable) {
    const fn = (iterableElement) => {
      const outputObject = {};
      const pairs = zipAll(propertyNames, iterableElement);
      for (const [propertyName, value] of pairs) {
        if (propertyName === undefined) continue
        outputObject[propertyName] = value;
      }
      return outputObject
    };
    // implement arrayToObject in terms of map
    return map(fn, iterable)
  };

  /**
   * Chunk every n items into an array, and output that array in the output sequence.
   *
   * @param {number} n - the number of items to group into each array.
   * @param {Iterable} iterable - the sequence of items to group
   * @return {Generator} for the chunked sequence
   * @example
   * const a = chunk(2, [0, 1, 2, 3, 4, 5, 6])
   * console.log([...a]) // prints [[0, 1], [2, 3], [4, 5], [6]]
   */
  const chunk = function * (n, iterable) {
    let accumulator = [];
    for (const item of iterable) {
      accumulator.push(item);
      if (accumulator.length === n) {
        yield accumulator;
        accumulator = [];
      }
    }
    if (accumulator.length > 0) {
      yield accumulator;
    }
  };

  /**
   * Keeps item from input sequence when fn(item) returns truthy. Remove items from input sequence when
   * fn(item) returns !truthy.
   *
   * @param {Function} fn - fn(item) returns truthy when item should be removed
   * @param {Iterable} iterable - the sequence to filter
   * @return {Generator} for the filtered sequence
   * @example
   * const isEvenNumber = x => x % 2 === 0
   * const a = filter(isEvenNumber, [0, 1, 2, 3, 4, 5, 6])
   * console.log([...a]) // prints even numbers [0, 2, 4, 6]
   */
  const filter = function * (fn, iterable) {
    for (const value of iterable) {
      if (fn(value) === true) {
        yield value;
      }
    }
  };

  // These two functions are used by flatten and flattenRecursive.
  // Generally accepted method of checking if something is a string.
  const isString = (item) => typeof item === 'string' || item instanceof String;
  // Generally accepted method of checking if something supports the Iterable protocol
  const isIterable = (item) => item && typeof item[Symbol.iterator] === 'function';

  /**
   * Flattens a sequence of items one level deep. It does not flatten strings, even
   * though they are iterable.
   *
   * @param {Iterable} iterable - the iterable sequence to flatten
   * @returns {Generator} for the flattened sequence
   * @example
   * const a = flatten([[0, 1], [2, 3], [4, 5], [6]])
   * console.log([...a]) // prints [0, 1, 2, 3, 4, 5, 6]
   */
  const flatten = function * (iterable) {
    for (const value of iterable) {
      if (!isString(value) && isIterable(value)) {
        yield * value;
      } else {
        yield value;
      }
    }
  };

  /**
   * Flattens a sequence by recursively returning items from each iterable in the sequence.
   * Does not flatten strings even though they are iterable.
   *
   * @param {Iterable} iterable - the sequence to flatten
   * @returns {Generator} for the flattened sequence
   * @example
   * const input = [0, [1, 2, 3], [[4, 5], [[[6, 7]], [8, 9], 10]], 11, 12]
   * const a = flattenRecursive(input)
   * console.log([...a]) // prints [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]
   */
  const flattenRecursive = function * (iterable) {
    for (const value of iterable) {
      if (!isString(value) && isIterable(value)) {
        yield * flattenRecursive(value);
      } else {
        yield value;
      }
    }
  };

  /**
   * Generates a sequence of items by calling fn(item) for each item.
   *
   * @param {Function} fn - fn(item) returns the output item
   * @param {Iterable} iterable - the sequence to map
   * @returns {Generator} for the mapped sequence
   * @example
   * const a = map(x => 2 * x, [0, 1, 2, 3])
   * console.log([...a]) // prints [0, 2, 4, 6]
   */
  const map = function * (fn, iterable) {
    for (const value of iterable) {
      yield fn(value);
    }
  };

  /**
   * Map the input sequence to the output sequence with a generator that maps one iterator to another.
   *
   * This method exists solely so that ChainableIterable supports chaining for an arbitrary generator function.
   *
   * @param {Function} generatorFunction - a function that returns an iterable object, and takes an iterable as a parameter.
   * Typically, this will be a generator function.
   * @param {Iterable} iterable - the input sequence
   * @returns {Generator} for the mapped sequence
   * @example
   * const fn = function * (iterable) {
   *   for (let x of iterable) {
   *     yield x * x
   *   }
   * }
   * const a = mapWith(fn, [0, 1, 2, 3])
   * console.log([...a]) // prints [0, 1, 4, 9]
   */
  const mapWith = function (generatorFunction, iterable) {
    return generatorFunction(iterable)
  };

  /**
   * Given a sequence of Arrays, output the nth element of each array as a sequence.
   *
   * @param {number} index - the index of the Array to output
   * @param {Iterable} iterable - the iterable to process
   * @returns {Generator} for the nth elements
   * @example
   * const input = [[0, 1], [2, 3], [4, 5]]
   * const a = nth(1, input)
   * console.log([...a]) // prints [1, 3, 5]
   */
  const nth = (index, iterable) => {
    return map(x => (index < 0) ? x[x.length + index] : x[index], iterable)
  };

  /**
   * Given a sequence of Objects, output the specified property of each Object as a sequence.
   *
   * @param {string} propertyname - the property to extract from each Object
   * @param {Iterable} iterable - the input sequence of Objects
   * @returns {Generator} for the plucked items
   * @example
   * const input = [{'a': 1, 'b': 2}, {'a': 3, 'b': 4}, {'a': 5, 'b': 6}]
   * const a = pluck('a', input)
   * console.log([...a]) // prints [1, 3, 5]
   */
  const pluck = (propertyname, iterable) => {
    return map(x => x[propertyname], iterable)
  };

  /**
   * Reject items when fn(item) returns truthy.
   *
   * @param {Function} fn - fn(item) returns truthy when item should be removed from output sequence
   * @param {Iterable} iterable - input sequence
   * @returns {Generator} for the non-rejected items
   * @example
   * const isEvenNumber = x => x % 2 === 0
   * const a = reject(isEvenNumber, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
   * console.log([...a]) // prints [1, 3, 5, 7, 9]
   */
  const reject = (fn, iterable) => {
    return filter(x => fn(x) !== true, iterable)
  };

  /**
   * Create an output sequence that is the first n items of the input sequence.
   *
   * @param {number} n - the number of items to take
   * @param {Iterable} iterable - the input sequence to take items from
   * @returns {Generator} for the first n items
   * @example
   * const a = take(5, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
   * console.log([...a]) // prints [0, 1, 2, 3, 4]
   */
  const take = function * (n, iterable) {
    const iterator = iterable[Symbol.iterator]();
    let taken = 0;
    let next;
    while (taken < n) {
      next = iterator.next();
      taken += 1;
      if (next.done === true) break
      yield next.value;
    }
  };

  /**
   * Output items from the input iterable until fn(item) returns !truthy.
   *
   * @param {Function} fn - fn(item) returns truthy to put item in the output sequence
   * @param {Iterable} iterable - input sequence
   * @returns {Generator} for the selected items
   * @example
   * const a = takeWhile(x => x != 4, [0, 1, 2, 3, 4, 5, 6])
   * console.log([...a]) // prints [0, 1, 2, 3]
   */
  const takeWhile = function * (fn, iterable) {
    for (const item of iterable) {
      if (fn(item) !== true) break
      yield item;
    }
  };

  /**
   * Pass the input sequence to the output sequence without change, but execute `fn(item)` for each
   * item in the sequence.
   *
   * @param {Function} fn - `fn(item)` is called for each item in the sequence
   * @param {Iterable} iterable - the input sequence
   * @returns {Generator} that is equivalent to the input iterable
   * @example
   * const a = tap(console.log, [1, 2, 3, 4, 5])
   * [...a] // prints [1, 2, 3, 4, 5]
   */
  const tap = (fn, iterable) => {
    return map(x => { fn(x); return x }, iterable)
  };

  var transforms = /*#__PURE__*/Object.freeze({
    arrayToObject: arrayToObject,
    chunk: chunk,
    filter: filter,
    flatten: flatten,
    flattenRecursive: flattenRecursive,
    map: map,
    mapWith: mapWith,
    nth: nth,
    pluck: pluck,
    reject: reject,
    take: take,
    takeWhile: takeWhile,
    tap: tap
  });

  /**
   * Executes function fn(item, index) for each item in the iterable sequence provided.
   *
   * @param {Function} fn - a function(item, index), where item is the current item in the sequence, and index
   * is the index of the current item.
   * @param {Iterable} iterable - the sequence of items to call fn(item, index) with.
   * @returns {undefined}
   * @example
   * const fn = (item, index) => console.log(`item - ${item}, index - ${index}`)
   * forEach(fn, [1, 2, 3]) // prints the following...
   * // item - 1, index - 0
   * // item - 2, index - 1
   * // item - 3, index - 2
   */
  const forEach = (fn, iterable) => {
    let index = 0;
    for (const item of iterable) {
      fn(item, index++);
    }
  };

  /**
   * The reduce() method executes a reducer function (that you provide)
   * on each element of the array, resulting in a single output value.
   *
   * @param {Function} fn - fn(accumulator, item) that returns the new accumulator value
   * @param {any} accumulator - the initial accumulator value
   * @param {Iterable} iterable - the sequence to execute fn over.
   * @returns {any} - the final accumulator value
   * @example
   * const add = (a, b) => a + b
   * const sum = reduce(add, 0, [0, 1, 2, 3, 4]) // sum === 10
   */
  const reduce = (fn, accumulator, iterable) => {
    for (const item of iterable) {
      accumulator = fn(accumulator, item);
    }
    return accumulator
  };

  /**
   * Creates an  Array from the items in iterable.
   *
   * @param {Iterable} iterable - the iterable to create the array from
   * @returns {Array} - the array
   * @example
   * const a = toArray(range(5)) // a is [0, 1, 2, 3, 4]
   */
  const toArray = (iterable) => Array.from(iterable);

  var reducers = /*#__PURE__*/Object.freeze({
    forEach: forEach,
    reduce: reduce,
    toArray: toArray
  });

  // Create an easy default chainable iterator builder for common use case
  const chainable = makeChainableIterable(generators, transforms, reducers);
  const ChainableIterable = chainable.ChainableIterable;

  exports.ChainableIterable = ChainableIterable;
  exports.chainable = chainable;
  exports.generators = generators;
  exports.makeChainableClass = makeChainableClass;
  exports.makeChainableIterable = makeChainableIterable;
  exports.reducers = reducers;
  exports.transforms = transforms;

  Object.defineProperty(exports, '__esModule', { value: true });

}));
//# sourceMappingURL=iterablefu.umd.js.map
